@@ -4,7 +4,7 @@
  *	  page utilities routines for the postgres inverted index access method.
  *
  *
- * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2013, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -112,8 +112,10 @@ ginFindLeafPage(GinBtree btree, GinBtreeStack *stack)
 				/* rightmost page */
 				break;
 
-			stack->buffer = ginStepRight(stack->buffer, btree->index, access);
 			stack->blkno = rightlink;
+			LockBuffer(stack->buffer, GIN_UNLOCK);
+			stack->buffer = ReleaseAndReadBuffer(stack->buffer, btree->index, stack->blkno);
+			LockBuffer(stack->buffer, access);
 			page = BufferGetPage(stack->buffer);
 		}
 
@@ -144,44 +146,6 @@ ginFindLeafPage(GinBtree btree, GinBtreeStack *stack)
 			stack->predictNumber = 1;
 		}
 	}
-
-	/* keep compiler happy */
-	return NULL;
-}
-
-/*
- * Step right from current page.
- *
- * The next page is locked first, before releasing the current page. This is
- * crucial to protect from concurrent page deletion (see comment in
- * ginDeletePage).
- */
-Buffer
-ginStepRight(Buffer buffer, Relation index, int lockmode)
-{
-	Buffer		nextbuffer;
-	Page		page = BufferGetPage(buffer);
-	bool		isLeaf = GinPageIsLeaf(page);
-	bool		isData = GinPageIsData(page);
-	BlockNumber	blkno = GinPageGetOpaque(page)->rightlink;
-
-	nextbuffer = ReadBuffer(index, blkno);
-	LockBuffer(nextbuffer, lockmode);
-	UnlockReleaseBuffer(buffer);
-
-	/* Sanity check that the page we stepped to is of similar kind. */
-	page = BufferGetPage(nextbuffer);
-	if (isLeaf != GinPageIsLeaf(page) || isData != GinPageIsData(page))
-		elog(ERROR, "right sibling of GIN page is of different type");
-
-	/*
-	 * Given the proper lock sequence above, we should never land on a
-	 * deleted page.
-	 */
-	if  (GinPageIsDeleted(page))
-		elog(ERROR, "right sibling of GIN page was deleted");
-
-	return nextbuffer;
 }
 
 void
@@ -209,7 +173,6 @@ void
 ginFindParents(GinBtree btree, GinBtreeStack *stack,
 			   BlockNumber rootBlkno)
 {
-
 	Page		page;
 	Buffer		buffer;
 	BlockNumber blkno,
@@ -272,12 +235,12 @@ ginFindParents(GinBtree btree, GinBtreeStack *stack,
 		while ((offset = btree->findChildPtr(btree, page, stack->blkno, InvalidOffsetNumber)) == InvalidOffsetNumber)
 		{
 			blkno = GinPageGetOpaque(page)->rightlink;
+			LockBuffer(buffer, GIN_UNLOCK);
+			ReleaseBuffer(buffer);
 			if (blkno == InvalidBlockNumber)
-			{
-				UnlockReleaseBuffer(buffer);
 				break;
-			}
-			buffer = ginStepRight(buffer, btree->index, GIN_EXCLUSIVE);
+			buffer = ReadBuffer(btree->index, blkno);
+			LockBuffer(buffer, GIN_EXCLUSIVE);
 			page = BufferGetPage(buffer);
 		}
 
@@ -308,20 +271,22 @@ ginFindParents(GinBtree btree, GinBtreeStack *stack,
 void
 ginInsertValue(GinBtree btree, GinBtreeStack *stack, GinStatsData *buildStats)
 {
-	GinBtreeStack *parent = stack;
-	BlockNumber rootBlkno = InvalidBuffer;
+	GinBtreeStack *parent;
+	BlockNumber rootBlkno;
 	Page		page,
 				rpage,
 				lpage;
 
-	/* remember root BlockNumber */
-	while (parent)
-	{
-		rootBlkno = parent->blkno;
+	/* extract root BlockNumber from stack */
+	Assert(stack != NULL);
+	parent = stack;
+	while (parent->parent)
 		parent = parent->parent;
-	}
+	rootBlkno = parent->blkno;
+	Assert(BlockNumberIsValid(rootBlkno));
 
-	while (stack)
+	/* this loop crawls up the stack until the insertion is complete */
+	for (;;)
 	{
 		XLogRecData *rdata;
 		BlockNumber savedRightLink;
@@ -342,7 +307,6 @@ ginInsertValue(GinBtree btree, GinBtreeStack *stack, GinStatsData *buildStats)
 
 				recptr = XLogInsert(RM_GIN_ID, XLOG_GIN_INSERT, rdata);
 				PageSetLSN(page, recptr);
-				PageSetTLI(page, ThisTimeLineID);
 			}
 
 			LockBuffer(stack->buffer, GIN_UNLOCK);
@@ -411,11 +375,8 @@ ginInsertValue(GinBtree btree, GinBtreeStack *stack, GinStatsData *buildStats)
 
 					recptr = XLogInsert(RM_GIN_ID, XLOG_GIN_SPLIT, rdata);
 					PageSetLSN(page, recptr);
-					PageSetTLI(page, ThisTimeLineID);
 					PageSetLSN(lpage, recptr);
-					PageSetTLI(lpage, ThisTimeLineID);
 					PageSetLSN(rpage, recptr);
-					PageSetTLI(rpage, ThisTimeLineID);
 				}
 
 				UnlockReleaseBuffer(rbuffer);
@@ -460,9 +421,7 @@ ginInsertValue(GinBtree btree, GinBtreeStack *stack, GinStatsData *buildStats)
 
 					recptr = XLogInsert(RM_GIN_ID, XLOG_GIN_SPLIT, rdata);
 					PageSetLSN(lpage, recptr);
-					PageSetTLI(lpage, ThisTimeLineID);
 					PageSetLSN(rpage, recptr);
-					PageSetTLI(rpage, ThisTimeLineID);
 				}
 				UnlockReleaseBuffer(rbuffer);
 				END_CRIT_SECTION();
@@ -480,21 +439,23 @@ ginInsertValue(GinBtree btree, GinBtreeStack *stack, GinStatsData *buildStats)
 		{
 			BlockNumber rightlink = GinPageGetOpaque(page)->rightlink;
 
+			LockBuffer(parent->buffer, GIN_UNLOCK);
+
 			if (rightlink == InvalidBlockNumber)
 			{
 				/*
 				 * rightmost page, but we don't find parent, we should use
 				 * plain search...
 				 */
-				LockBuffer(parent->buffer, GIN_UNLOCK);
 				ginFindParents(btree, stack, rootBlkno);
 				parent = stack->parent;
-				page = BufferGetPage(parent->buffer);
+				Assert(parent != NULL);
 				break;
 			}
 
-			parent->buffer = ginStepRight(parent->buffer, btree->index, GIN_EXCLUSIVE);
 			parent->blkno = rightlink;
+			parent->buffer = ReleaseAndReadBuffer(parent->buffer, btree->index, parent->blkno);
+			LockBuffer(parent->buffer, GIN_EXCLUSIVE);
 			page = BufferGetPage(parent->buffer);
 		}
 
