@@ -3,7 +3,7 @@
  * rewriteDefine.c
  *	  routines for defining a rewrite rule
  *
- * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2013, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
@@ -15,8 +15,13 @@
 #include "postgres.h"
 
 #include "access/heapam.h"
+#include "access/htup_details.h"
+#include "access/multixact.h"
+#include "access/transam.h"
+#include "access/xact.h"
 #include "catalog/catalog.h"
 #include "catalog/dependency.h"
+#include "catalog/heap.h"
 #include "catalog/indexing.h"
 #include "catalog/namespace.h"
 #include "catalog/objectaccess.h"
@@ -177,8 +182,7 @@ InsertRule(char *rulname,
 	}
 
 	/* Post creation hook for new rule */
-	InvokeObjectAccessHook(OAT_POST_CREATE,
-						   RewriteRelationId, rewriteObjectId, 0, NULL);
+	InvokeObjectPostCreateHook(RewriteRelationId, rewriteObjectId, 0);
 
 	heap_close(pg_rewrite_desc, RowExclusiveLock);
 
@@ -189,7 +193,7 @@ InsertRule(char *rulname,
  * DefineRule
  *		Execute a CREATE RULE command.
  */
-void
+Oid
 DefineRule(RuleStmt *stmt, const char *queryString)
 {
 	List	   *actions;
@@ -200,19 +204,19 @@ DefineRule(RuleStmt *stmt, const char *queryString)
 	transformRuleStmt(stmt, queryString, &actions, &whereClause);
 
 	/*
-	 * Find and lock the relation.  Lock level should match
+	 * Find and lock the relation.	Lock level should match
 	 * DefineQueryRewrite.
 	 */
 	relId = RangeVarGetRelid(stmt->relation, AccessExclusiveLock, false);
 
 	/* ... and execute */
-	DefineQueryRewrite(stmt->rulename,
-					   relId,
-					   whereClause,
-					   stmt->event,
-					   stmt->instead,
-					   stmt->replace,
-					   actions);
+	return DefineQueryRewrite(stmt->rulename,
+							  relId,
+							  whereClause,
+							  stmt->event,
+							  stmt->instead,
+							  stmt->replace,
+							  actions);
 }
 
 
@@ -223,7 +227,7 @@ DefineRule(RuleStmt *stmt, const char *queryString)
  * This is essentially the same as DefineRule() except that the rule's
  * action and qual have already been passed through parse analysis.
  */
-void
+Oid
 DefineQueryRewrite(char *rulename,
 				   Oid event_relid,
 				   Node *event_qual,
@@ -237,6 +241,7 @@ DefineQueryRewrite(char *rulename,
 	ListCell   *l;
 	Query	   *query;
 	bool		RelisBecomingView = false;
+	Oid			ruleId = InvalidOid;
 
 	/*
 	 * If we are installing an ON SELECT rule, we had better grab
@@ -252,8 +257,11 @@ DefineQueryRewrite(char *rulename,
 
 	/*
 	 * Verify relation is of a type that rules can sensibly be applied to.
+	 * Internal callers can target materialized views, but transformRuleStmt()
+	 * blocks them for users.  Don't mention them in the error message.
 	 */
 	if (event_relation->rd_rel->relkind != RELKIND_RELATION &&
+		event_relation->rd_rel->relkind != RELKIND_MATVIEW &&
 		event_relation->rd_rel->relkind != RELKIND_VIEW)
 		ereport(ERROR,
 				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
@@ -402,13 +410,14 @@ DefineQueryRewrite(char *rulename,
 		 *
 		 * If so, check that the relation is empty because the storage for the
 		 * relation is going to be deleted.  Also insist that the rel not have
-		 * any triggers, indexes, or child tables.  (Note: these tests are too
+		 * any triggers, indexes, or child tables.	(Note: these tests are too
 		 * strict, because they will reject relations that once had such but
 		 * don't anymore.  But we don't really care, because this whole
 		 * business of converting relations to views is just a kluge to allow
-		 * loading ancient pg_dump files.)
+		 * dump/reload of views that participate in circular dependencies.)
 		 */
-		if (event_relation->rd_rel->relkind != RELKIND_VIEW)
+		if (event_relation->rd_rel->relkind != RELKIND_VIEW &&
+			event_relation->rd_rel->relkind != RELKIND_MATVIEW)
 		{
 			HeapScanDesc scanDesc;
 
@@ -487,38 +496,119 @@ DefineQueryRewrite(char *rulename,
 	/* discard rule if it's null action and not INSTEAD; it's a no-op */
 	if (action != NIL || is_instead)
 	{
-		InsertRule(rulename,
-				   event_type,
-				   event_relid,
-				   event_attno,
-				   is_instead,
-				   event_qual,
-				   action,
-				   replace);
+		ruleId = InsertRule(rulename,
+							event_type,
+							event_relid,
+							event_attno,
+							is_instead,
+							event_qual,
+							action,
+							replace);
 
 		/*
-		 * Set pg_class 'relhasrules' field TRUE for event relation. If
-		 * appropriate, also modify the 'relkind' field to show that the
-		 * relation is now a view.
+		 * Set pg_class 'relhasrules' field TRUE for event relation.
 		 *
 		 * Important side effect: an SI notice is broadcast to force all
 		 * backends (including me!) to update relcache entries with the new
 		 * rule.
 		 */
-		SetRelationRuleStatus(event_relid, true, RelisBecomingView);
+		SetRelationRuleStatus(event_relid, true);
 	}
 
-	/*
-	 * IF the relation is becoming a view, delete the storage files associated
-	 * with it.  NB: we had better have AccessExclusiveLock to do this ...
+	/* ---------------------------------------------------------------------
+	 * If the relation is becoming a view:
+	 * - delete the associated storage files
+	 * - get rid of any system attributes in pg_attribute; a view shouldn't
+	 *	 have any of those
+	 * - remove the toast table; there is no need for it anymore, and its
+	 *	 presence would make vacuum slightly more complicated
+	 * - set relkind to RELKIND_VIEW, and adjust other pg_class fields
+	 *	 to be appropriate for a view
 	 *
-	 * XXX what about getting rid of its TOAST table?  For now, we don't.
+	 * NB: we had better have AccessExclusiveLock to do this ...
+	 * ---------------------------------------------------------------------
 	 */
 	if (RelisBecomingView)
+	{
+		Relation	relationRelation;
+		Oid			toastrelid;
+		HeapTuple	classTup;
+		Form_pg_class classForm;
+
+		relationRelation = heap_open(RelationRelationId, RowExclusiveLock);
+		toastrelid = event_relation->rd_rel->reltoastrelid;
+
+		/* drop storage while table still looks like a table  */
 		RelationDropStorage(event_relation);
+		DeleteSystemAttributeTuples(event_relid);
+
+		/*
+		 * Drop the toast table if any.  (This won't take care of updating the
+		 * toast fields in the relation's own pg_class entry; we handle that
+		 * below.)
+		 */
+		if (OidIsValid(toastrelid))
+		{
+			ObjectAddress toastobject;
+
+			/*
+			 * Delete the dependency of the toast relation on the main
+			 * relation so we can drop the former without dropping the latter.
+			 */
+			deleteDependencyRecordsFor(RelationRelationId, toastrelid,
+									   false);
+
+			/* Make deletion of dependency record visible */
+			CommandCounterIncrement();
+
+			/* Now drop toast table, including its index */
+			toastobject.classId = RelationRelationId;
+			toastobject.objectId = toastrelid;
+			toastobject.objectSubId = 0;
+			performDeletion(&toastobject, DROP_RESTRICT,
+							PERFORM_DELETION_INTERNAL);
+		}
+
+		/*
+		 * SetRelationRuleStatus may have updated the pg_class row, so we must
+		 * advance the command counter before trying to update it again.
+		 */
+		CommandCounterIncrement();
+
+		/*
+		 * Fix pg_class entry to look like a normal view's, including setting
+		 * the correct relkind and removal of reltoastrelid/reltoastidxid of
+		 * the toast table we potentially removed above.
+		 */
+		classTup = SearchSysCacheCopy1(RELOID, ObjectIdGetDatum(event_relid));
+		if (!HeapTupleIsValid(classTup))
+			elog(ERROR, "cache lookup failed for relation %u", event_relid);
+		classForm = (Form_pg_class) GETSTRUCT(classTup);
+
+		classForm->reltablespace = InvalidOid;
+		classForm->relpages = 0;
+		classForm->reltuples = 0;
+		classForm->relallvisible = 0;
+		classForm->reltoastrelid = InvalidOid;
+		classForm->reltoastidxid = InvalidOid;
+		classForm->relhasindex = false;
+		classForm->relkind = RELKIND_VIEW;
+		classForm->relhasoids = false;
+		classForm->relhaspkey = false;
+		classForm->relfrozenxid = InvalidTransactionId;
+		classForm->relminmxid = InvalidMultiXactId;
+
+		simple_heap_update(relationRelation, &classTup->t_self, classTup);
+		CatalogUpdateIndexes(relationRelation, classTup);
+
+		heap_freetuple(classTup);
+		heap_close(relationRelation, RowExclusiveLock);
+	}
 
 	/* Close rel, but keep lock till commit... */
 	heap_close(event_relation, NoLock);
+
+	return ruleId;
 }
 
 /*
@@ -539,7 +629,6 @@ checkRuleResultList(List *targetList, TupleDesc resultDesc, bool isSelect)
 	foreach(tllist, targetList)
 	{
 		TargetEntry *tle = (TargetEntry *) lfirst(tllist);
-		Oid			tletypid;
 		int32		tletypmod;
 		Form_pg_attribute attr;
 		char	   *attname;
@@ -571,32 +660,19 @@ checkRuleResultList(List *targetList, TupleDesc resultDesc, bool isSelect)
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 					 errmsg("cannot convert relation containing dropped columns to view")));
 
-		/* Check name match if required; no need for two error texts here */
 		if (isSelect && strcmp(tle->resname, attname) != 0)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
-					 errmsg("SELECT rule's target entry %d has different column name from \"%s\"",
-							i, attname),
-					 errdetail("SELECT target entry is named \"%s\".",
-							   tle->resname)));
+					 errmsg("SELECT rule's target entry %d has different column name from \"%s\"", i, attname)));
 
-		/* Check type match. */
-		tletypid = exprType((Node *) tle->expr);
-		if (attr->atttypid != tletypid)
+		if (attr->atttypid != exprType((Node *) tle->expr))
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
 					 isSelect ?
 					 errmsg("SELECT rule's target entry %d has different type from column \"%s\"",
 							i, attname) :
 					 errmsg("RETURNING list's entry %d has different type from column \"%s\"",
-							i, attname),
-					 isSelect ?
-					 errdetail("SELECT target entry has type %s, but column has type %s.",
-							   format_type_be(tletypid),
-							   format_type_be(attr->atttypid)) :
-					 errdetail("RETURNING list entry has type %s, but column has type %s.",
-							   format_type_be(tletypid),
-							   format_type_be(attr->atttypid))));
+							i, attname)));
 
 		/*
 		 * Allow typmods to be different only if one of them is -1, ie,
@@ -613,16 +689,7 @@ checkRuleResultList(List *targetList, TupleDesc resultDesc, bool isSelect)
 					 errmsg("SELECT rule's target entry %d has different size from column \"%s\"",
 							i, attname) :
 					 errmsg("RETURNING list's entry %d has different size from column \"%s\"",
-							i, attname),
-					 isSelect ?
-					 errdetail("SELECT target entry has type %s, but column has type %s.",
-							   format_type_with_typemod(tletypid, tletypmod),
-							   format_type_with_typemod(attr->atttypid,
-														attr->atttypmod)) :
-					 errdetail("RETURNING list entry has type %s, but column has type %s.",
-							   format_type_with_typemod(tletypid, tletypmod),
-							   format_type_with_typemod(attr->atttypid,
-														attr->atttypmod))));
+							i, attname)));
 	}
 
 	if (i != resultDesc->natts)
@@ -641,7 +708,7 @@ checkRuleResultList(List *targetList, TupleDesc resultDesc, bool isSelect)
  * Note: for a view (ON SELECT rule), the checkAsUser field of the OLD
  * RTE entry will be overridden when the view rule is expanded, and the
  * checkAsUser field of the NEW entry is irrelevant because that entry's
- * requiredPerms bits will always be zero.  However, for other types of rules
+ * requiredPerms bits will always be zero.	However, for other types of rules
  * it's important to set these fields to match the rule owner.  So we just set
  * them always.
  */
@@ -750,6 +817,9 @@ EnableDisableRule(Relation rel, const char *rulename,
 		changed = true;
 	}
 
+	InvokeObjectPostAlterHook(RewriteRelationId,
+							  HeapTupleGetOid(ruletup), 0);
+
 	heap_freetuple(ruletup);
 	heap_close(pg_rewrite_desc, RowExclusiveLock);
 
@@ -764,37 +834,98 @@ EnableDisableRule(Relation rel, const char *rulename,
 
 
 /*
- * Rename an existing rewrite rule.
- *
- * This is unused code at the moment.  Note that it lacks a permissions check.
+ * Perform permissions and integrity checks before acquiring a relation lock.
  */
-#ifdef NOT_USED
-void
-RenameRewriteRule(Oid owningRel, const char *oldName,
+static void
+RangeVarCallbackForRenameRule(const RangeVar *rv, Oid relid, Oid oldrelid,
+							  void *arg)
+{
+	HeapTuple	tuple;
+	Form_pg_class form;
+
+	tuple = SearchSysCache1(RELOID, ObjectIdGetDatum(relid));
+	if (!HeapTupleIsValid(tuple))
+		return;					/* concurrently dropped */
+	form = (Form_pg_class) GETSTRUCT(tuple);
+
+	/* only tables and views can have rules */
+	if (form->relkind != RELKIND_RELATION && form->relkind != RELKIND_VIEW)
+		ereport(ERROR,
+				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+				 errmsg("\"%s\" is not a table or view", rv->relname)));
+
+	if (!allowSystemTableMods && IsSystemClass(form))
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+				 errmsg("permission denied: \"%s\" is a system catalog",
+						rv->relname)));
+
+	/* you must own the table to rename one of its rules */
+	if (!pg_class_ownercheck(relid, GetUserId()))
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS, rv->relname);
+
+	ReleaseSysCache(tuple);
+}
+
+/*
+ * Rename an existing rewrite rule.
+ */
+Oid
+RenameRewriteRule(RangeVar *relation, const char *oldName,
 				  const char *newName)
 {
+	Oid			relid;
+	Relation	targetrel;
 	Relation	pg_rewrite_desc;
 	HeapTuple	ruletup;
+	Form_pg_rewrite ruleform;
+	Oid			ruleOid;
 
+	/*
+	 * Look up name, check permissions, and acquire lock (which we will NOT
+	 * release until end of transaction).
+	 */
+	relid = RangeVarGetRelidExtended(relation, AccessExclusiveLock,
+									 false, false,
+									 RangeVarCallbackForRenameRule,
+									 NULL);
+
+	/* Have lock already, so just need to build relcache entry. */
+	targetrel = relation_open(relid, NoLock);
+
+	/* Prepare to modify pg_rewrite */
 	pg_rewrite_desc = heap_open(RewriteRelationId, RowExclusiveLock);
 
+	/* Fetch the rule's entry (it had better exist) */
 	ruletup = SearchSysCacheCopy2(RULERELNAME,
-								  ObjectIdGetDatum(owningRel),
+								  ObjectIdGetDatum(relid),
 								  PointerGetDatum(oldName));
 	if (!HeapTupleIsValid(ruletup))
 		ereport(ERROR,
 				(errcode(ERRCODE_UNDEFINED_OBJECT),
 				 errmsg("rule \"%s\" for relation \"%s\" does not exist",
-						oldName, get_rel_name(owningRel))));
+						oldName, RelationGetRelationName(targetrel))));
+	ruleform = (Form_pg_rewrite) GETSTRUCT(ruletup);
+	ruleOid = HeapTupleGetOid(ruletup);
 
-	/* should not already exist */
-	if (IsDefinedRewriteRule(owningRel, newName))
+	/* rule with the new name should not already exist */
+	if (IsDefinedRewriteRule(relid, newName))
 		ereport(ERROR,
 				(errcode(ERRCODE_DUPLICATE_OBJECT),
 				 errmsg("rule \"%s\" for relation \"%s\" already exists",
-						newName, get_rel_name(owningRel))));
+						newName, RelationGetRelationName(targetrel))));
 
-	namestrcpy(&(((Form_pg_rewrite) GETSTRUCT(ruletup))->rulename), newName);
+	/*
+	 * We disallow renaming ON SELECT rules, because they should always be
+	 * named "_RETURN".
+	 */
+	if (ruleform->ev_type == CMD_SELECT + '0')
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+				 errmsg("renaming an ON SELECT rule is not allowed")));
+
+	/* OK, do the update */
+	namestrcpy(&(ruleform->rulename), newName);
 
 	simple_heap_update(pg_rewrite_desc, &ruletup->t_self, ruletup);
 
@@ -803,6 +934,18 @@ RenameRewriteRule(Oid owningRel, const char *oldName,
 
 	heap_freetuple(ruletup);
 	heap_close(pg_rewrite_desc, RowExclusiveLock);
-}
 
-#endif
+	/*
+	 * Invalidate relation's relcache entry so that other backends (and this
+	 * one too!) are sent SI message to make them rebuild relcache entries.
+	 * (Ideally this should happen automatically...)
+	 */
+	CacheInvalidateRelcache(targetrel);
+
+	/*
+	 * Close rel, but keep exclusive lock!
+	 */
+	relation_close(targetrel, NoLock);
+
+	return ruleOid;
+}
